import os

from jinja2 import Template
from tqdm import tqdm

from .adapters import get_adapter
from .generate_base_test import make_test
from .runner import MultiThreadRunner
from .templater import get_template
from .utils import DEFAULT_DBT_PROFILE_PATH, list_yaml, read_dbt_profile, read_source_yaml, write_file

template = get_template("base_model.sql")


def make_model_sql(*args, **kwargs):
    sql = template.render(*args, **kwargs)
    return sql


def generate_base_from_source(adapter, output_folder, source):
    source_identifier = source.get("schema", source["name"])

    for table in source["tables"]:
        table_name = table["name"]
        table_identifier = table.get("identifier", table_name)

        # Identifier is name used in database
        columns = adapter.list_columns(source_identifier, table_identifier)

        # While, name is used in dbt
        sql_content = make_model_sql({"source_name": source["name"], "table_name": table_name, "columns": columns})
        sql_path = os.path.join(output_folder, f"stg__{table_name}.sql")
        write_file(sql_path, sql_content)


def generate_base_model(profile_path, output_folder, source_path, profile_name="default", target="dev", threads=None):
    os.makedirs(output_folder, exist_ok=True)

    profile = read_dbt_profile(profile_path, profile_name=profile_name, target=target)

    bm_tasks = {}
    bt_tasks = {}
    for path in list_yaml(source_path):
        for source in read_source_yaml(path):
            adapter = get_adapter(profile, database=source["database"])

            # Make subfolder for base models
            sub_folder = os.path.join(output_folder, source["name"])
            os.makedirs(sub_folder, exist_ok=True)

            # Base model task
            bm_tasks[source["name"]] = {"adapter": adapter, "output_folder": sub_folder, "source": source}

            # Base test task
            bt_tasks[source["name"]] = {
                "adapter": adapter,
                "source": source,
                "output_path": os.path.join(sub_folder, "schema.yml"),
            }

    bm_runner = MultiThreadRunner(generate_base_from_source, bm_tasks, threads=threads)
    for result in tqdm(bm_runner.run(), total=len(bm_tasks)):
        # print(f"Generated base models of source: {result['task_name']}")
        pass
    print("Done generating base models")

    bt_runner = MultiThreadRunner(make_test, bt_tasks, threads=threads)
    for result in tqdm(bt_runner.run(), total=len(bt_tasks)):
        # print(f"Generated base tests of source: {result['task_name']}")
        pass
    print("Done generating base tests")


def run(args):
    if args.template:
        global template
        print(f"Use custom template at {args.template}")
        with open(args.template, "r", encoding="utf-8") as file:
            template = Template(file.read())

    generate_base_model(
        profile_path=args.profile_path,
        output_folder=args.output_folder,
        source_path=args.source_path,
        profile_name=args.profile_name,
        target=args.target,
        threads=args.threads,
    )


def config_parser(parser):
    parser.set_defaults(func=run)
    parser.add_argument("source_path", type=str, help="Path to dbt source YAML.")
    parser.add_argument(
        "output_folder",
        type=str,
        help="Folder to write base models.",
    )
    parser.add_argument(
        "--profile-path",
        type=str,
        default=DEFAULT_DBT_PROFILE_PATH,
        help=("Path to dbt profile YAML." f" Default is {DEFAULT_DBT_PROFILE_PATH}"),
    )
    parser.add_argument("--profile-name", type=str, default="default", help="Dbt profile name. Default is `default`.")
    parser.add_argument("--target", type=str, default="dev", help="Dbt profile target. Default is `dev`.")
    parser.add_argument(
        "--threads",
        type=int,
        default=None,
        help="Max threads. Default is your machine number of threads.",
    )
    parser.add_argument("--template", type=str, default=None, help="Path to custom Jinja template.")



import os
from typing import Any, Mapping, Optional

from .adapters import BaseAdapter, BigQueryAdapter, SnowflakeAdapter, get_adapter
from .runner import Runner
from .templater import Templater
from .utils import DEFAULT_DBT_PROFILE_PATH, list_yaml, read_dbt_profile, read_source_yaml, write_file

template = Templater().load_template("base_tests.yml")


class UnsupportedException(Exception):
    pass


def get_snowflake_pkeys(adapter: SnowflakeAdapter, source: Mapping[str, Any]) -> Mapping[str, Any]:
    pkeys = {}
    for table in source["tables"]:
        schema = source.get("schema", source["name"])
        query = f'describe table "{source["database"]}"."{schema.upper()}"."{table["name"].upper()}"'
        cols = []
        for row in adapter.execute(query):
            if row["primary key"] == "Y":
                cols.append(row["name"].lower())
        pkeys[table["name"]] = cols
    return pkeys


def get_bigquery_pkeys__stitch(adapter: BigQueryAdapter, source: Mapping[str, Any]) -> Mapping[str, Any]:
    schema = source.get("schema", source["name"])
    pkeys = {}

    if "_sdc_primary_keys" in list(adapter.list_tables(schema)):
        info_df = adapter.execute_pandas(
            f"select table_name, column_name from `{source['database']}.{schema}._sdc_primary_keys`;"
        )
        for _, row in info_df.iterrows():
            cols = pkeys.setdefault(row["table_name"], [])
            cols.append(row["column_name"])
    else:
        print(f"WARNING: There is no Stitch primary key table in the dataset {schema}. Tests will be empty.")
    return pkeys


def get_pkeys(adapter: BaseAdapter, source: Mapping[str, Any]) -> Mapping[str, Any]:
    if isinstance(adapter, BigQueryAdapter):
        return get_bigquery_pkeys__stitch(adapter, source)
    elif isinstance(adapter, SnowflakeAdapter):
        return get_snowflake_pkeys(adapter, source)
    else:
        raise UnsupportedException(f"Adapter name {adapter.name} is unsupported")


def make_test(adapter: BaseAdapter, source: Mapping[str, Any], output_path: str, **kwargs) -> None:
    pkeys = get_pkeys(adapter, source)
    table_info = [
        {"name": table, "description": "", "pkeys": " || '-' || ".join(pkeys), "extra": kwargs}
        for table, pkeys in pkeys.items()
    ]
    test_yaml = template.render(tables=table_info)
    write_file(output_path, test_yaml)


def generate_base_tests(
    profile_path: str,
    source_path: str,
    output_folder: str,
    profile_name: Optional[str] = "default",
    target: Optional[str] = "dev",
    extra: Optional[Mapping[str, Any]] = None,
):
    extra = extra or {}
    profile = read_dbt_profile(profile_path, profile_name, target)

    tasks = {}
    for path in list_yaml(source_path):
        for source in read_source_yaml(path):
            adapter = get_adapter(profile, database=source["database"])
            tasks[source["name"]] = {
                "adapter": adapter,
                "source": source,
                "output_path": os.path.join(output_folder, f'{source["name"]}.yml'),
            }

    runner = Runner(make_test, tasks, **extra)

    for result in runner.run():
        print(f"Source {result['task_name']}")


def run(args):
    generate_base_tests(
        profile_path=args.profile_path,
        source_path=args.source_path,
        output_folder=args.output_folder,
        profile_name=args.profile_name,
        target=args.target,
        extra=None,
    )


def config_parser(parser):
    parser.set_defaults(func=run)
    parser.add_argument("source_path", type=str, help="Path to dbt source YAML.")
    parser.add_argument(
        "output_folder",
        type=str,
        help="Folder to write base models.",
    )
    parser.add_argument(
        "--profile-path",
        type=str,
        default=DEFAULT_DBT_PROFILE_PATH,
        help=("Path to dbt profile YAML." f" Default is {DEFAULT_DBT_PROFILE_PATH}"),
    )
    parser.add_argument("--profile-name", type=str, default="default", help="Dbt profile name. Default is `default`.")
    parser.add_argument("--target", type=str, default="dev", help="Dbt profile target. Default is `dev`.")



import os
from pathlib import Path

import yaml

DEFAULT_DBT_PROFILE_PATH = os.path.join(str(Path.home()), ".dbt", "profiles.yml")


def read_yaml(path):
    with open(path, "r", encoding="utf-8") as file:
        return yaml.load(file, Loader=yaml.FullLoader)


def read_dbt_profile(file_path, profile_name="default", target="dev"):
    profiles = read_yaml(file_path)

    if profile_name not in profiles:
        raise ValueError(f"Profile `{profile_name}` not found.")

    profile = profiles[profile_name]

    if target not in profile["outputs"]:
        raise ValueError(f"Target `{target}` not found.")

    return profile["outputs"][target]


def write_file(file_path, text):
    with open(file_path, "w", encoding="utf-8") as file:
        file.write(text)


def read_source_yaml(path):
    source_config = read_yaml(path)

    if "sources" not in source_config:
        raise ValueError("Source YAML malformed. Missing `sources`.")

    return source_config["sources"]


def list_yaml(path):
    if os.path.isdir(path):
        return [os.path.join(path, child) for child in os.listdir(path) if os.path.isfile(os.path.join(path, child))]
    elif os.path.isfile(path):
        return [path]
    else:
        raise ValueError(f"Path {path} is not a file or directory")



try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources

from . import templates


def print_default_template(args):
    print(pkg_resources.read_text(templates, "base_model.sql"))


def config_parser(parser):
    parser.set_defaults(func=print_default_template)
